@@ -4,8 +4,17 @@ Halohalo API Module for managing the Segmentation Platform.
 
 import requests
 import json
-import time
 from pyspark.sql.functions import *
+from pymongo import *
+import time
+import pyspark
+import math
+from pyspark.sql.types import *
+from pyspark.sql.functions import col
+import concurrent
+import concurrent.futures
+from pyspark.context import SparkContext
+from pyspark.sql.session import SparkSession
 
 
 class HalohaloAPI:
@@ -415,3 +424,223 @@ class HalohaloAPI:
             print(res["response"])
 
         return res
+
+    def segment_bulk_writer(
+        self,
+        input_df,
+        target_segment,
+        target_collection,
+        mongo_aws_access_key_id,
+        mongo_aws_secret_access_key,
+    ):
+        # initialize spark
+        sc = SparkContext.getOrCreate()
+        spark = SparkSession(sc)
+        # filter input_df using target segment
+        filtered_source_df = input_df.filter(
+            col("source_segment_name") == target_segment
+        ).select(col("source_user_id"), col("source_segment_name"))
+
+        # filter duplicate ID
+        source_df = filtered_source_df.dropDuplicates(["source_user_id"])
+
+        try:
+            existing_df = (
+                spark.read.format("com.mongodb.spark.sql.DefaultSource")
+                .option("uri", self.uri)
+                .option("collection", target_collection)
+                .load()
+                .filter(col("segment_name") == target_segment)
+                .select(
+                    col("_id"), col("user_id"), col("segment_name"), col("is_member")
+                )
+            )
+        except Exception as e:
+            raise Exception("ERROR:", e)
+
+        flags = []
+        # --no existing records in db for target segment, add all as new records
+        if existing_df.count() == 0:
+            if int(existing_df.count()) == 0:
+                raise Exception(
+                    f"Target segment {target_segment} does not exist in {target_collection}"
+                )
+        else:
+            flags = ["activate", "new", "deactivate"]
+
+        def ingest_parallelize_spark_df_multithread(
+            rows,
+            flag,
+            mongo_aws_access_key_id,
+            mongo_aws_secret_access_key,
+            cluster,
+            database,
+            target_collection,
+        ):
+
+            # os installation
+            import os
+
+            os.system("pip install -U kumuniverse")
+
+            from kumuniverse.mongodb import Mongo
+
+            operations = []
+            # can be passed as flag
+            collection = target_collection
+            mongo = Mongo(mongo_aws_access_key_id, mongo_aws_secret_access_key, cluster)
+
+            rows = list(rows)
+            if len(rows) > 0:
+                for row in rows:
+
+                    if flag == "new":
+                        operations.append(InsertOne(row.asDict()))
+                    else:
+                        row_dict = row.asDict()
+                        item_id = row_dict["user_id"]
+                        is_member = row_dict["is_member"]
+                        segment_name = row_dict["segment_name"]
+                        updated_at = row_dict["updated_at"]
+                        # update flag -can be either activate or deactivate
+                        operate = UpdateOne(
+                            {"user_id": item_id, "segment_name": segment_name},
+                            {
+                                "$set": {
+                                    "is_member": is_member,
+                                    "updated_at": updated_at,
+                                }
+                            },
+                        )
+                        operations.append(operate)
+
+                    # Send once every 1000 in batch
+                    if len(operations) == 1000:
+                        try:
+                            mongo.client[database][collection].bulk_write(
+                                operations, ordered=False
+                            )
+                            operations = []
+                        except Exception as e:
+                            print(str(e))
+
+                if len(operations) > 0:
+                    try:
+                        mongo.client[database][collection].bulk_write(
+                            operations, ordered=False
+                        )
+                        operations = []
+                    except Exception as e:
+                        print(str(e))
+
+        def process_df(flag):
+            join_conds = [
+                existing_df.user_id == source_df.source_user_id,
+                existing_df.segment_name == source_df.source_segment_name,
+            ]
+            if flag == "new":
+
+                # ---- new
+                join_conds = [
+                    existing_df.user_id == source_df.source_user_id,
+                    existing_df.segment_name == source_df.source_segment_name,
+                ]
+                new_user_df = (
+                    source_df.join(existing_df, join_conds, "left_anti")
+                    .withColumn("created_at", lit(int(time.time())))
+                    .withColumn("updated_at", lit(int(time.time())))
+                    .withColumn("is_member", lit(True))
+                    .select(
+                        col("source_user_id").alias("user_id"),
+                        col("source_segment_name").alias("segment_name"),
+                        col("is_member"),
+                        col("created_at"),
+                        col("updated_at"),
+                    )
+                )
+                input_df = new_user_df
+
+            elif flag == "activate":
+
+                # convert to is_member = True
+                update_to_true = (
+                    source_df.join(existing_df, join_conds, "inner")
+                    .filter(existing_df.is_member == False)
+                    .drop(col("is_member"))
+                    .withColumn("is_member", lit(True))
+                    .select(
+                        col("_id"),
+                        col("source_user_id").alias("user_id"),
+                        col("source_segment_name").alias("segment_name"),
+                        col("is_member"),
+                    )
+                    .withColumn("updated_at", lit(int(time.time())))
+                )
+                input_df = update_to_true
+            else:
+                # convert to is_member = False
+                update_to_false = (
+                    source_df.join(existing_df, join_conds, "inner")
+                    .filter(existing_df.is_member == True)
+                    .drop(col("is_member"))
+                    .withColumn("is_member", lit(False))
+                    .select(
+                        col("_id"),
+                        col("source_user_id").alias("user_id"),
+                        col("source_segment_name").alias("segment_name"),
+                        col("is_member"),
+                    )
+                    .withColumn("updated_at", lit(int(time.time())))
+                )
+
+                input_df = update_to_false
+
+            # --partition logic----#
+
+            if input_df.count() > 0:
+                if input_df.count() == 1:
+                    num_of_records_each_partition = 1
+                    partition_count = 2
+
+                elif input_df.count() <= 200000:  # if record < 200k
+                    num_of_records_each_partition = int(
+                        math.floor(input_df.count() / 2)
+                    )
+                    partition_count = int(
+                        math.floor(input_df.count() / num_of_records_each_partition)
+                    )
+
+                else:
+                    partition_count = int(math.floor(input_df.count() / 100000))
+                    num_of_records_each_partition = int(
+                        math.floor(input_df.count() / partition_count)
+                    )
+                record_count = input_df.count()
+                repartitioned = input_df.repartition(partition_count)
+
+                try:
+
+                    repartitioned.foreachPartition(
+                        lambda rows: ingest_parallelize_spark_df_multithread(
+                            rows,
+                            flag,
+                            mongo_aws_access_key_id,
+                            mongo_aws_secret_access_key,
+                            self.cluster,
+                            self.database,
+                            target_collection,
+                        )
+                    )
+                    return f"Value of input_df count: {record_count} for flag {flag}. Partition count: {partition_count} and record for each partition:  {num_of_records_each_partition}"
+                except Exception as e:
+                    raise Exception(e)
+            else:
+                return f"No value for flag {flag}."
+
+        with concurrent.futures.ThreadPoolExecutor(max_workers=len(flags)) as executor:
+            futures = [executor.submit(process_df, flag) for flag in flags]
+            for future in concurrent.futures.as_completed(futures):
+                if future.exception() is not None:
+                    print("future exception: ", str(future.exception()))
+                else:
+                    print(future.result())
